"""
Pixel manipulation methods.

Pixel manipulation is used in N2V and similar algorithm to replace the value of
masked pixels.
"""

from typing import Optional, Tuple

import numpy as np

from .struct_mask_parameters import StructMaskParameters


def _apply_struct_mask(
    patch: np.ndarray,
    coords: np.ndarray,
    struct_params: StructMaskParameters,
    rng: Optional[np.random.Generator] = None,
) -> np.ndarray:
    """Apply structN2V masks to patch.

    Each point in `coords` corresponds to the center of a mask, masks are paremeterized
    by `struct_params` and pixels in the mask (with respect to `coords`) are replaced by
    a random value.

    Note that the structN2V mask is applied in 2D at the coordinates given by `coords`.

    Parameters
    ----------
    patch : np.ndarray
        Patch to be manipulated, 2D or 3D.
    coords : np.ndarray
        Coordinates of the ROI(subpatch) centers.
    struct_params : StructMaskParameters
        Parameters for the structN2V mask (axis and span).
    rng : np.random.Generator or None
        Random number generator.

    Returns
    -------
    np.ndarray
        Patch with the structN2V mask applied.
    """
    if rng is None:
        rng = np.random.default_rng()

    # relative axis
    moving_axis = -1 - struct_params.axis

    # Create a mask array
    mask = np.expand_dims(
        np.ones(struct_params.span), axis=list(range(len(patch.shape) - 1))
    )  # (1, 1, span) or (1, span)

    # Move the moving axis to the correct position
    # i.e. the axis along which the coordinates should change
    mask = np.moveaxis(mask, -1, moving_axis)
    center = np.array(mask.shape) // 2

    # Mark the center
    mask[tuple(center.T)] = 0

    # displacements from center
    dx = np.indices(mask.shape)[:, mask == 1] - center[:, None]

    # combine all coords (ndim, npts,) with all displacements (ncoords,ndim,)
    mix = dx.T[..., None] + coords.T[None]
    mix = mix.transpose([1, 0, 2]).reshape([mask.ndim, -1]).T

    # delete entries that are out of bounds
    mix = np.delete(mix, mix[:, moving_axis] < 0, axis=0)

    max_bound = patch.shape[moving_axis] - 1
    mix = np.delete(mix, mix[:, moving_axis] > max_bound, axis=0)

    # replace neighbouring pixels with random values from flat dist
    patch[tuple(mix.T)] = rng.uniform(patch.min(), patch.max(), size=mix.shape[0])

    return patch


def _odd_jitter_func(step: float, rng: np.random.Generator) -> np.ndarray:
    """
    Randomly sample a jitter to be applied to the masking grid.

    This is done to account for cases where the step size is not an integer.

    Parameters
    ----------
    step : float
        Step size of the grid, output of np.linspace.
    rng : np.random.Generator
        Random number generator.

    Returns
    -------
    np.ndarray
        Array of random jitter to be added to the grid.
    """
    # Define the random jitter to be added to the grid
    odd_jitter = np.where(np.floor(step) == step, 0, rng.integers(0, 2))

    # Round the step size to the nearest integer depending on the jitter
    return np.floor(step) if odd_jitter == 0 else np.ceil(step)


def _get_stratified_coords(
    mask_pixel_perc: float,
    shape: Tuple[int, ...],
    rng: Optional[np.random.Generator] = None,
) -> np.ndarray:
    """
    Generate coordinates of the pixels to mask.

    Randomly selects the coordinates of the pixels to mask in a stratified way, i.e.
    the distance between masked pixels is approximately the same.

    Parameters
    ----------
    mask_pixel_perc : float
        Actual (quasi) percentage of masked pixels across the whole image. Used in
        calculating the distance between masked pixels across each axis.
    shape : Tuple[int, ...]
        Shape of the input patch.
    rng : np.random.Generator or None
        Random number generator.

    Returns
    -------
    np.ndarray
        Array of coordinates of the masked pixels.
    """
    if len(shape) < 2 or len(shape) > 3:
        raise ValueError(
            "Calculating coordinates is only possible for 2D and 3D patches"
        )

    if rng is None:
        rng = np.random.default_rng()

    mask_pixel_distance = np.round((100 / mask_pixel_perc) ** (1 / len(shape))).astype(
        np.int32
    )

    # Define a grid of coordinates for each axis in the input patch and the step size
    pixel_coords = []
    steps = []
    for axis_size in shape:
        # make sure axis size is evenly divisible by box size
        num_pixels = int(np.ceil(axis_size / mask_pixel_distance))
        axis_pixel_coords, step = np.linspace(
            0, axis_size, num_pixels, dtype=np.int32, endpoint=False, retstep=True
        )
        # explain
        pixel_coords.append(axis_pixel_coords.T)
        steps.append(step)

    # Create a meshgrid of coordinates for each axis in the input patch
    coordinate_grid_list = np.meshgrid(*pixel_coords)
    coordinate_grid = np.array(coordinate_grid_list).reshape(len(shape), -1).T

    grid_random_increment = rng.integers(
        _odd_jitter_func(float(max(steps)), rng)  # type: ignore
        * np.ones_like(coordinate_grid).astype(np.int32)
        - 1,
        size=coordinate_grid.shape,
        endpoint=True,
    )
    coordinate_grid += grid_random_increment
    coordinate_grid = np.clip(coordinate_grid, 0, np.array(shape) - 1)
    return coordinate_grid


def _create_subpatch_center_mask(
    subpatch: np.ndarray, center_coords: np.ndarray
) -> np.ndarray:
    """Create a mask with the center of the subpatch masked.

    Parameters
    ----------
    subpatch : np.ndarray
        Subpatch to be manipulated.
    center_coords : np.ndarray
        Coordinates of the original center before possible crop.

    Returns
    -------
    np.ndarray
        Mask with the center of the subpatch masked.
    """
    mask = np.ones(subpatch.shape)
    mask[tuple(center_coords)] = 0
    return np.ma.make_mask(mask)  # type: ignore


def _create_subpatch_struct_mask(
    subpatch: np.ndarray, center_coords: np.ndarray, struct_params: StructMaskParameters
) -> np.ndarray:
    """Create a structN2V mask for the subpatch.

    Parameters
    ----------
    subpatch : np.ndarray
        Subpatch to be manipulated.
    center_coords : np.ndarray
        Coordinates of the original center before possible crop.
    struct_params : StructMaskParameters
        Parameters for the structN2V mask (axis and span).

    Returns
    -------
    np.ndarray
        StructN2V mask for the subpatch.
    """
    # Create a mask with the center of the subpatch masked
    mask_placeholder = np.ones(subpatch.shape)

    # reshape to move the struct axis to the first position
    mask_reshaped = np.moveaxis(mask_placeholder, struct_params.axis, 0)

    # create the mask index for the struct axis
    mask_index = slice(
        max(0, center_coords.take(struct_params.axis) - (struct_params.span - 1) // 2),
        min(
            1 + center_coords.take(struct_params.axis) + (struct_params.span - 1) // 2,
            subpatch.shape[struct_params.axis],
        ),
    )
    mask_reshaped[struct_params.axis][mask_index] = 0

    # reshape back to the original shape
    mask = np.moveaxis(mask_reshaped, 0, struct_params.axis)

    return np.ma.make_mask(mask)  # type: ignore


def uniform_manipulate(
    patch: np.ndarray,
    mask_pixel_percentage: float,
    subpatch_size: int = 11,
    remove_center: bool = True,
    struct_params: Optional[StructMaskParameters] = None,
    rng: Optional[np.random.Generator] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Manipulate pixels by replacing them with a neighbor values.

    Manipulated pixels are selected unformly selected in a subpatch, away from a grid
    with an approximate uniform probability to be selected across the whole patch.
    If `struct_params` is not None, an additional structN2V mask is applied to the
    data, replacing the pixels in the mask with random values (excluding the pixel
    already manipulated).

    Parameters
    ----------
    patch : np.ndarray
        Image patch, 2D or 3D, shape (y, x) or (z, y, x).
    mask_pixel_percentage : float
        Approximate percentage of pixels to be masked.
    subpatch_size : int
        Size of the subpatch the new pixel value is sampled from, by default 11.
    remove_center : bool
        Whether to remove the center pixel from the subpatch, by default False.
    struct_params : StructMaskParameters or None
        Parameters for the structN2V mask (axis and span).
    rng : np.random.Generator or None
        Random number generator.

    Returns
    -------
    Tuple[np.ndarray]
        Tuple containing the manipulated patch and the corresponding mask.
    """
    if rng is None:
        rng = np.random.default_rng()

    # Get the coordinates of the pixels to be replaced
    transformed_patch = patch.copy()

    subpatch_centers = _get_stratified_coords(mask_pixel_percentage, patch.shape, rng)

    # Generate coordinate grid for subpatch
    roi_span_full = np.arange(
        -np.floor(subpatch_size / 2), np.ceil(subpatch_size / 2)
    ).astype(np.int32)

    # Remove the center pixel from the grid if needed
    roi_span = roi_span_full[roi_span_full != 0] if remove_center else roi_span_full

    # Randomly select coordinates from the grid
    random_increment = rng.choice(roi_span, size=subpatch_centers.shape)

    # Clip the coordinates to the patch size
    replacement_coords = np.clip(
        subpatch_centers + random_increment,
        0,
        [patch.shape[i] - 1 for i in range(len(patch.shape))],
    )

    # Get the replacement pixels from all subpatchs
    replacement_pixels = patch[tuple(replacement_coords.T.tolist())]

    # Replace the original pixels with the replacement pixels
    transformed_patch[tuple(subpatch_centers.T.tolist())] = replacement_pixels
    mask = np.where(transformed_patch != patch, 1, 0).astype(np.uint8)

    if struct_params is not None:
        transformed_patch = _apply_struct_mask(
            transformed_patch, subpatch_centers, struct_params
        )

    return (
        transformed_patch,
        mask,
    )


def median_manipulate(
    patch: np.ndarray,
    mask_pixel_percentage: float,
    subpatch_size: int = 11,
    struct_params: Optional[StructMaskParameters] = None,
    rng: Optional[np.random.Generator] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Manipulate pixels by replacing them with the median of their surrounding subpatch.

    N2V2 version, manipulated pixels are selected randomly away from a grid with an
    approximate uniform probability to be selected across the whole patch.

    If `struct_params` is not None, an additional structN2V mask is applied to the data,
    replacing the pixels in the mask with random values (excluding the pixel already
    manipulated).

    Parameters
    ----------
    patch : np.ndarray
        Image patch, 2D or 3D, shape (y, x) or (z, y, x).
    mask_pixel_percentage : floar
        Approximate percentage of pixels to be masked.
    subpatch_size : int
        Size of the subpatch the new pixel value is sampled from, by default 11.
    struct_params : StructMaskParameters or None, optional
        Parameters for the structN2V mask (axis and span).
    rng : np.random.Generator or None, optional
        Random number generato, by default None.

    Returns
    -------
    Tuple[np.ndarray]
           Tuple containing the manipulated patch, the original patch and the mask.
    """
    if rng is None:
        rng = np.random.default_rng()

    transformed_patch = patch.copy()

    # Get the coordinates of the pixels to be replaced
    subpatch_centers = _get_stratified_coords(mask_pixel_percentage, patch.shape, rng)

    # Generate coordinate grid for subpatch
    roi_span = np.array(
        [-np.floor(subpatch_size / 2), np.ceil(subpatch_size / 2)]
    ).astype(np.int32)

    subpatch_crops_span_full = subpatch_centers[np.newaxis, ...].T + roi_span

    # Dimensions n dims, n centers, (min, max)
    subpatch_crops_span_clipped = np.clip(
        subpatch_crops_span_full,
        a_min=np.zeros_like(patch.shape)[:, np.newaxis, np.newaxis],
        a_max=np.array(patch.shape)[:, np.newaxis, np.newaxis],
    )

    for idx in range(subpatch_crops_span_clipped.shape[1]):
        subpatch_coords = subpatch_crops_span_clipped[:, idx, ...]
        idxs = [
            slice(x[0], x[1]) if x[1] - x[0] > 0 else slice(0, 1)
            for x in subpatch_coords
        ]
        subpatch = patch[tuple(idxs)]
        subpatch_center_adjusted = subpatch_centers[idx] - subpatch_coords[:, 0]

        if struct_params is None:
            subpatch_mask = _create_subpatch_center_mask(
                subpatch, subpatch_center_adjusted
            )
        else:
            subpatch_mask = _create_subpatch_struct_mask(
                subpatch, subpatch_center_adjusted, struct_params
            )
        transformed_patch[tuple(subpatch_centers[idx])] = np.median(
            subpatch[subpatch_mask]
        )

    mask = np.where(transformed_patch != patch, 1, 0).astype(np.uint8)

    if struct_params is not None:
        transformed_patch = _apply_struct_mask(
            transformed_patch, subpatch_centers, struct_params
        )

    return (
        transformed_patch,
        mask,
    )



import numpy as np
import pytest

from careamics.dataset.tiling import extract_tiles
from careamics.prediction_utils import convert_outputs


@pytest.mark.parametrize("n_samples", [1, 2])
@pytest.mark.parametrize("batch_size", [1, 2])
def test_convert_outputs_tiled(ordered_array, batch_size, n_samples):
    """Test conversion of output for when prediction is tiled"""
    # --- simulate outputs from trainer.predict
    tile_size = (4, 4)
    overlaps = (2, 2)
    arr = ordered_array((n_samples, 1, 16, 16))
    all_tiles = list(extract_tiles(arr, tile_size, overlaps))
    # combine into batches to match output of trainer.predict
    prediction_batches = []
    for i in range(0, len(all_tiles), batch_size):
        tiles = np.concatenate(
            [tile[np.newaxis] for tile, _ in all_tiles[i : i + batch_size]], axis=0
        )
        tile_infos = [tile_info for _, tile_info in all_tiles[i : i + batch_size]]
        prediction_batches.append((tiles, tile_infos))

    predictions = convert_outputs(prediction_batches, tiled=True)
    assert np.array_equal(np.stack(predictions, axis=0).squeeze(), arr.squeeze())


@pytest.mark.parametrize("batch_size, n_samples", [(1, 1), (1, 2), (2, 2)])
def test_convert_outputs_not_tiled(ordered_array, batch_size, n_samples):
    """Test conversion of output for when prediction is not tiled"""
    # --- simulate outputs from trainer.predict
    prediction_batches = [
        ordered_array((batch_size, 1, 16, 16)) for _ in range(n_samples // batch_size)
    ]
    predictions = convert_outputs(prediction_batches, tiled=False)
    assert np.array_equal(
        # stack predictions because there is no S axis
        # squeeze to remove singleton S or C axes
        np.stack(predictions, axis=0).squeeze(),
        np.concatenate(prediction_batches, axis=0).squeeze(),
    )



"""Methods for Loss Computation."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, Optional, Union

import numpy as np
import torch

from careamics.losses.lvae.loss_utils import free_bits_kl, get_kl_weight
from careamics.models.lvae.likelihoods import (
    GaussianLikelihood,
    LikelihoodModule,
    NoiseModelLikelihood,
)

if TYPE_CHECKING:
    from careamics.config import LVAELossConfig

Likelihood = Union[LikelihoodModule, GaussianLikelihood, NoiseModelLikelihood]


def get_reconstruction_loss(
    reconstruction: torch.Tensor,
    target: torch.Tensor,
    likelihood_obj: Likelihood,
) -> dict[str, torch.Tensor]:
    """Compute the reconstruction loss (negative log-likelihood).

    Parameters
    ----------
    reconstruction: torch.Tensor
        The output of the LVAE decoder. Shape is (B, C, [Z], Y, X), where C is the
        number of output channels (e.g., 1 in HDN, >1 in muSplit/denoiSplit).
    target: torch.Tensor
        The target image used to compute the reconstruction loss. Shape is
        (B, C, [Z], Y, X), where C is the number of output channels
        (e.g., 1 in HDN, >1 in muSplit/denoiSplit).
    likelihood_obj: Likelihood
        The likelihood object used to compute the reconstruction loss.

    Returns
    -------
    torch.Tensor
        The recontruction loss (negative log-likelihood).
    """
    # Compute Log likelihood
    ll, _ = likelihood_obj(reconstruction, target)  # shape: (B, C, [Z], Y, X)
    return -1 * ll.mean()


def _reconstruction_loss_musplit_denoisplit(
    predictions: Union[torch.Tensor, tuple[torch.Tensor, torch.Tensor]],
    targets: torch.Tensor,
    nm_likelihood: NoiseModelLikelihood,
    gaussian_likelihood: GaussianLikelihood,
    nm_weight: float,
    gaussian_weight: float,
) -> torch.Tensor:
    """Compute the reconstruction loss for muSplit-denoiSplit loss.

    The resulting loss is a weighted mean of the noise model likelihood and the
    Gaussian likelihood.

    Parameters
    ----------
    predictions : torch.Tensor
        The output of the LVAE decoder. Shape is (B, C, [Z], Y, X), or
        (B, 2*C, [Z], Y, X), where C is the number of output channels,
        and the factor of 2 is for the case of predicted log-variance.
    targets : torch.Tensor
        The target image used to compute the reconstruction loss. Shape is
        (B, C, [Z], Y, X), where C is the number of output channels
        (e.g., 1 in HDN, >1 in muSplit/denoiSplit).
    nm_likelihood : NoiseModelLikelihood
        A `NoiseModelLikelihood` object used to compute the noise model likelihood.
    gaussian_likelihood : GaussianLikelihood
        A `GaussianLikelihood` object used to compute the Gaussian likelihood.
    nm_weight : float
        The weight for the noise model likelihood.
    gaussian_weight : float
        The weight for the Gaussian likelihood.

    Returns
    -------
    recons_loss : torch.Tensor
        The reconstruction loss. Shape is (1, ).
    """
    if predictions.shape[1] == 2 * targets.shape[1]:
        # predictions contain both mean and log-variance
        pred_mean, _ = predictions.chunk(2, dim=1)
    else:
        pred_mean = predictions

    recons_loss_nm = get_reconstruction_loss(
        reconstruction=pred_mean, target=targets, likelihood_obj=nm_likelihood
    )

    recons_loss_gm = get_reconstruction_loss(
        reconstruction=predictions,
        target=targets,
        likelihood_obj=gaussian_likelihood,
    )

    recons_loss = nm_weight * recons_loss_nm + gaussian_weight * recons_loss_gm
    return recons_loss


def get_kl_divergence_loss(
    kl_type: Literal["kl", "kl_restricted"],
    topdown_data: dict[str, torch.Tensor],
    rescaling: Literal["latent_dim", "image_dim"],
    aggregation: Literal["mean", "sum"],
    free_bits_coeff: float,
    img_shape: Optional[tuple[int]] = None,
) -> torch.Tensor:
    """Compute the KL divergence loss.

    NOTE: Description of `rescaling` methods:
    - If "latent_dim", the KL-loss values are rescaled w.r.t. the latent space
    dimensions (spatial + number of channels, i.e., (C, [Z], Y, X)). In this way they
    have the same magnitude across layers.
    - If "image_dim", the KL-loss values are rescaled w.r.t. the input image spatial
    dimensions. In this way, the lower layers have a larger KL-loss value compared to
    the higher layers, since the latent space and hence the KL tensor has more entries.
    Specifically, at hierarchy `i`, the total KL loss is larger by a factor (128/i**2).

    NOTE: the type of `aggregation` determines the magnitude of the KL-loss. Clearly,
    "sum" aggregation results in a larger KL-loss value compared to "mean" by a factor
    of `n_layers`.

    NOTE: recall that sample-wise KL is obtained by summing over all dimensions,
    including Z. Also recall that in current 3D implementation of LVAE, no downsampling
    is done on Z. Therefore, to avoid emphasizing KL loss too much, we divide it
    by the Z dimension of input image in every case.

    Parameters
    ----------
    kl_type : Literal["kl", "kl_restricted"]
        The type of KL divergence loss to compute.
    topdown_data : dict[str, torch.Tensor]
        A dictionary containing information computed for each layer during the top-down
        pass. The dictionary must include the following keys:
        - "kl": The KL-loss values for each layer. Shape of each tensor is (B,).
        - "z": The sampled latents for each layer. Shape of each tensor is
        (B, layers, `z_dims[i]`, H, W).
    rescaling : Literal["latent_dim", "image_dim"]
        The rescaling method used for the KL-loss values. If "latent_dim", the KL-loss
        values are rescaled w.r.t. the latent space dimensions (spatial + number of
        channels, i.e., (C, [Z], Y, X)). If "image_dim", the KL-loss values are
        rescaled w.r.t. the input image spatial dimensions.
    aggregation : Literal["mean", "sum"]
        The aggregation method used to combine the KL-loss values across layers. If
        "mean", the KL-loss values are averaged across layers. If "sum", the KL-loss
        values are summed across layers.
    free_bits_coeff : float
        The free bits coefficient used for the KL-loss computation.
    img_shape : Optional[tuple[int]]
        The shape of the input image to the LVAE model. Shape is ([Z], Y, X).

    Returns
    -------
    kl_loss : torch.Tensor
        The KL divergence loss. Shape is (1, ).
    """
    kl = torch.cat(
        [kl_layer.unsqueeze(1) for kl_layer in topdown_data[kl_type]],
        dim=1,
    )  # shape: (B, n_layers)

    # Apply free bits (& batch average)
    kl = free_bits_kl(kl, free_bits_coeff)  # shape: (n_layers,)

    # In 3D case, rescale by Z dim
    # TODO If we have downsampling in Z dimension, then this needs to change.
    if len(img_shape) == 3:
        kl = kl / img_shape[0]

    # Rescaling
    if rescaling == "latent_dim":
        for i in range(len(kl)):
            latent_dim = topdown_data["z"][i].shape[1:]
            norm_factor = np.prod(latent_dim)
            kl[i] = kl[i] / norm_factor
    elif rescaling == "image_dim":
        kl = kl / np.prod(img_shape[-2:])

    # Aggregation
    if aggregation == "mean":
        kl = kl.mean()  # shape: (1,)
    elif aggregation == "sum":
        kl = kl.sum()  # shape: (1,)

    return kl


def _get_kl_divergence_loss_musplit(
    topdown_data: dict[str, torch.Tensor],
    img_shape: tuple[int],
    kl_type: Literal["kl", "kl_restricted"],
) -> torch.Tensor:
    """Compute the KL divergence loss for muSplit.

    Parameters
    ----------
    topdown_data : dict[str, torch.Tensor]
        A dictionary containing information computed for each layer during the top-down
        pass. The dictionary must include the following keys:
        - "kl": The KL-loss values for each layer. Shape of each tensor is (B,).
        - "z": The sampled latents for each layer. Shape of each tensor is
        (B, layers, `z_dims[i]`, H, W).
    img_shape : tuple[int]
        The shape of the input image to the LVAE model. Shape is ([Z], Y, X).
    kl_type : Literal["kl", "kl_restricted"]
        The type of KL divergence loss to compute.

    Returns
    -------
    kl_loss : torch.Tensor
        The KL divergence loss for the muSplit case. Shape is (1, ).
    """
    return get_kl_divergence_loss(
        kl_type="kl",  # TODO: hardcoded, deal in future PR
        topdown_data=topdown_data,
        rescaling="latent_dim",
        aggregation="mean",
        free_bits_coeff=0.0,
        img_shape=img_shape,
    )


def _get_kl_divergence_loss_denoisplit(
    topdown_data: dict[str, torch.Tensor],
    img_shape: tuple[int],
    kl_type: Literal["kl", "kl_restricted"],
) -> torch.Tensor:
    """Compute the KL divergence loss for denoiSplit.

    Parameters
    ----------
    topdown_data : dict[str, torch.Tensor]
        A dictionary containing information computed for each layer during the top-down
        pass. The dictionary must include the following keys:
        - "kl": The KL-loss values for each layer. Shape of each tensor is (B,).
        - "z": The sampled latents for each layer. Shape of each tensor is
        (B, layers, `z_dims[i]`, H, W).
    img_shape : tuple[int]
        The shape of the input image to the LVAE model. Shape is ([Z], Y, X).
    kl_type : Literal["kl", "kl_restricted"]
        The type of KL divergence loss to compute.

    Returns
    -------
    kl_loss : torch.Tensor
        The KL divergence loss for the denoiSplit case. Shape is (1, ).
    """
    return get_kl_divergence_loss(
        kl_type=kl_type,
        topdown_data=topdown_data,
        rescaling="image_dim",
        aggregation="sum",
        free_bits_coeff=1.0,
        img_shape=img_shape,
    )


# TODO: @melisande-c suggested to refactor this as a class (see PR #208)
# - loss computation happens by calling the `__call__` method
# - `__init__` method initializes the loss parameters now contained in
# the `LVAELossParameters` class
# NOTE: same for the other loss functions
def musplit_loss(
    model_outputs: tuple[torch.Tensor, dict[str, Any]],
    targets: torch.Tensor,
    config: LVAELossConfig,
    gaussian_likelihood: Optional[GaussianLikelihood],
    noise_model_likelihood: Optional[NoiseModelLikelihood] = None,  # TODO: ugly
) -> Optional[dict[str, torch.Tensor]]:
    """Loss function for muSplit.

    Parameters
    ----------
    model_outputs : tuple[torch.Tensor, dict[str, Any]]
        Tuple containing the model predictions (shape is (B, `target_ch`, [Z], Y, X))
        and the top-down layer data (e.g., sampled latents, KL-loss values, etc.).
    targets : torch.Tensor
        The target image used to compute the reconstruction loss. Shape is
        (B, `target_ch`, [Z], Y, X).
    config : LVAELossConfig
        The config for loss function (e.g., KL hyperparameters, likelihood module,
        noise model, etc.).
    gaussian_likelihood : GaussianLikelihood
        The Gaussian likelihood object.
    noise_model_likelihood : Optional[NoiseModelLikelihood]
        The noise model likelihood object. Not used here.

    Returns
    -------
    output : Optional[dict[str, torch.Tensor]]
        A dictionary containing the overall loss `["loss"]`, the reconstruction loss
        `["reconstruction_loss"]`, and the KL divergence loss `["kl_loss"]`.
    """
    assert gaussian_likelihood is not None

    predictions, td_data = model_outputs

    # Reconstruction loss computation
    recons_loss = config.reconstruction_weight * get_reconstruction_loss(
        reconstruction=predictions,
        target=targets,
        likelihood_obj=gaussian_likelihood,
    )
    if torch.isnan(recons_loss).any():
        recons_loss = 0.0

    # KL loss computation
    kl_weight = get_kl_weight(
        config.kl_params.annealing,
        config.kl_params.start,
        config.kl_params.annealtime,
        config.kl_weight,
        config.kl_params.current_epoch,
    )
    kl_loss = (
        _get_kl_divergence_loss_musplit(
            topdown_data=td_data,
            img_shape=targets.shape[2:],
            kl_type=config.kl_params.loss_type,
        )
        * kl_weight
    )

    net_loss = recons_loss + kl_loss
    output = {
        "loss": net_loss,
        "reconstruction_loss": (
            recons_loss.detach()
            if isinstance(recons_loss, torch.Tensor)
            else recons_loss
        ),
        "kl_loss": kl_loss.detach(),
    }
    # https://github.com/openai/vdvae/blob/main/train.py#L26
    if torch.isnan(net_loss).any():
        return None

    return output


def denoisplit_loss(
    model_outputs: tuple[torch.Tensor, dict[str, Any]],
    targets: torch.Tensor,
    config: LVAELossConfig,
    gaussian_likelihood: Optional[GaussianLikelihood] = None,
    noise_model_likelihood: Optional[NoiseModelLikelihood] = None,
) -> Optional[dict[str, torch.Tensor]]:
    """Loss function for DenoiSplit.

    Parameters
    ----------
    model_outputs : tuple[torch.Tensor, dict[str, Any]]
        Tuple containing the model predictions (shape is (B, `target_ch`, [Z], Y, X))
        and the top-down layer data (e.g., sampled latents, KL-loss values, etc.).
    targets : torch.Tensor
        The target image used to compute the reconstruction loss. Shape is
        (B, `target_ch`, [Z], Y, X).
    config : LVAELossConfig
        The config for loss function containing all loss hyperparameters.
    gaussian_likelihood : GaussianLikelihood
        The Gaussian likelihood object.
    noise_model_likelihood : NoiseModelLikelihood
        The noise model likelihood object.

    Returns
    -------
    output : Optional[dict[str, torch.Tensor]]
        A dictionary containing the overall loss `["loss"]`, the reconstruction loss
        `["reconstruction_loss"]`, and the KL divergence loss `["kl_loss"]`.
    """
    assert noise_model_likelihood is not None

    predictions, td_data = model_outputs

    # Reconstruction loss computation
    recons_loss = config.reconstruction_weight * get_reconstruction_loss(
        reconstruction=predictions,
        target=targets,
        likelihood_obj=noise_model_likelihood,
    )
    if torch.isnan(recons_loss).any():
        recons_loss = 0.0

    # KL loss computation
    kl_weight = get_kl_weight(
        config.kl_params.annealing,
        config.kl_params.start,
        config.kl_params.annealtime,
        config.kl_weight,
        config.kl_params.current_epoch,
    )
    kl_loss = (
        _get_kl_divergence_loss_denoisplit(
            topdown_data=td_data,
            img_shape=targets.shape[2:],
            kl_type=config.kl_params.loss_type,
        )
        * kl_weight
    )

    net_loss = recons_loss + kl_loss
    output = {
        "loss": net_loss,
        "reconstruction_loss": (
            recons_loss.detach()
            if isinstance(recons_loss, torch.Tensor)
            else recons_loss
        ),
        "kl_loss": kl_loss.detach(),
    }
    # https://github.com/openai/vdvae/blob/main/train.py#L26
    if torch.isnan(net_loss).any():
        return None

    return output


def denoisplit_musplit_loss(
    model_outputs: tuple[torch.Tensor, dict[str, Any]],
    targets: torch.Tensor,
    config: LVAELossConfig,
    gaussian_likelihood: GaussianLikelihood,
    noise_model_likelihood: NoiseModelLikelihood,
) -> Optional[dict[str, torch.Tensor]]:
    """Loss function for DenoiSplit.

    Parameters
    ----------
    model_outputs : tuple[torch.Tensor, dict[str, Any]]
        Tuple containing the model predictions (shape is (B, `target_ch`, [Z], Y, X))
        and the top-down layer data (e.g., sampled latents, KL-loss values, etc.).
    targets : torch.Tensor
        The target image used to compute the reconstruction loss. Shape is
        (B, `target_ch`, [Z], Y, X).
    config : LVAELossConfig
        The config for loss function containing all loss hyperparameters.
    gaussian_likelihood : GaussianLikelihood
        The Gaussian likelihood object.
    noise_model_likelihood : NoiseModelLikelihood
        The noise model likelihood object.

    Returns
    -------
    output : Optional[dict[str, torch.Tensor]]
        A dictionary containing the overall loss `["loss"]`, the reconstruction loss
        `["reconstruction_loss"]`, and the KL divergence loss `["kl_loss"]`.
    """
    predictions, td_data = model_outputs

    # Reconstruction loss computation
    recons_loss = _reconstruction_loss_musplit_denoisplit(
        predictions=predictions,
        targets=targets,
        nm_likelihood=noise_model_likelihood,
        gaussian_likelihood=gaussian_likelihood,
        nm_weight=config.denoisplit_weight,
        gaussian_weight=config.musplit_weight,
    )
    if torch.isnan(recons_loss).any():
        recons_loss = 0.0

    # KL loss computation
    # NOTE: 'kl' key stands for the 'kl_samplewise' key in the TopDownLayer class.
    # The different naming comes from `top_down_pass()` method in the LadderVAE.
    denoisplit_kl = _get_kl_divergence_loss_denoisplit(
        topdown_data=td_data,
        img_shape=targets.shape[2:],
        kl_type=config.kl_params.loss_type,
    )
    musplit_kl = _get_kl_divergence_loss_musplit(
        topdown_data=td_data,
        img_shape=targets.shape[2:],
        kl_type=config.kl_params.loss_type,
    )
    kl_loss = (
        config.denoisplit_weight * denoisplit_kl + config.musplit_weight * musplit_kl
    )
    # TODO `kl_weight` is hardcoded (???)
    kl_loss = config.kl_weight * kl_loss

    net_loss = recons_loss + kl_loss
    output = {
        "loss": net_loss,
        "reconstruction_loss": (
            recons_loss.detach()
            if isinstance(recons_loss, torch.Tensor)
            else recons_loss
        ),
        "kl_loss": kl_loss.detach(),
    }
    # https://github.com/openai/vdvae/blob/main/train.py#L26
    if torch.isnan(net_loss).any():
        return None

    return output



"""Pydantic model representing the metadata of a prediction tile."""

from __future__ import annotations

from typing import Annotated

from annotated_types import Len
from pydantic import BaseModel, ConfigDict

DimTuple = Annotated[tuple, Len(min_length=3, max_length=4)]


class TileInformation(BaseModel):
    """
    Pydantic model containing tile information.

    This model is used to represent the information required to stitch back a tile into
    a larger image. It is used throughout the prediction pipeline of CAREamics.

    Array shape should be C(Z)YX, where Z is an optional dimensions.
    """

    model_config = ConfigDict(validate_default=True)

    array_shape: DimTuple  # TODO: find a way to add custom error message?
    """Shape of the original (untiled) array."""

    last_tile: bool = False
    """Whether this tile is the last one of the array."""

    overlap_crop_coords: tuple[tuple[int, ...], ...]
    """Inner coordinates of the tile where to crop the prediction in order to stitch
    it back into the original image."""

    stitch_coords: tuple[tuple[int, ...], ...]
    """Coordinates in the original image where to stitch the cropped tile back."""

    sample_id: int
    """Sample ID of the tile."""

    # TODO: Test that ZYX axes are not singleton ?

    def __eq__(self, other_tile: object):
        """Check if two tile information objects are equal.

        Parameters
        ----------
        other_tile : object
            Tile information object to compare with.

        Returns
        -------
        bool
            Whether the two tile information objects are equal.
        """
        if not isinstance(other_tile, TileInformation):
            return NotImplemented

        return (
            self.array_shape == other_tile.array_shape
            and self.last_tile == other_tile.last_tile
            and self.overlap_crop_coords == other_tile.overlap_crop_coords
            and self.stitch_coords == other_tile.stitch_coords
            and self.sample_id == other_tile.sample_id
        )



from careamics.config.transformations import XYRandomRotate90Model
from careamics.transforms import XYRandomRotate90


def test_comptatibility_with_transform():
    """Test that the model allows instantiating a transform."""
    model = XYRandomRotate90Model(p=0.3)

    # instantiate transform
    XYRandomRotate90(**model.model_dump())



"""References for the CAREamics algorithms."""

from bioimageio.spec.generic.v0_3 import CiteEntry

N2VRef = CiteEntry(
    text='Krull, A., Buchholz, T.O. and Jug, F., 2019. "Noise2Void - Learning '
    'denoising from single noisy images". In Proceedings of the IEEE/CVF '
    "conference on computer vision and pattern recognition (pp. 2129-2137).",
    doi="10.1109/cvpr.2019.00223",
)

N2V2Ref = CiteEntry(
    text="Höck, E., Buchholz, T.O., Brachmann, A., Jug, F. and Freytag, A., "
    '2022. "N2V2 - Fixing Noise2Void checkerboard artifacts with modified '
    'sampling strategies and a tweaked network architecture". In European '
    "Conference on Computer Vision (pp. 503-518).",
    doi="10.1007/978-3-031-25069-9_33",
)

StructN2VRef = CiteEntry(
    text="Broaddus, C., Krull, A., Weigert, M., Schmidt, U. and Myers, G., 2020."
    '"Removing structured noise with self-supervised blind-spot '
    'networks". In 2020 IEEE 17th International Symposium on Biomedical '
    "Imaging (ISBI) (pp. 159-163).",
    doi="10.1109/isbi45749.2020.9098336",
)

N2NRef = CiteEntry(
    text="Lehtinen, J., Munkberg, J., Hasselgren, J., Laine, S., Karras, T., "
    'Aittala, M. and Aila, T., 2018. "Noise2Noise: Learning image restoration '
    'without clean data". arXiv preprint arXiv:1803.04189.',
    doi="10.48550/arXiv.1803.04189",
)

CARERef = CiteEntry(
    text='Weigert, Martin, et al. "Content-aware image restoration: pushing the '
    'limits of fluorescence microscopy." Nature methods 15.12 (2018): 1090-1097.',
    doi="10.1038/s41592-018-0216-7",
)



"""Utility functions for the CAREamics CLI."""

from typing import Optional, Tuple


def handle_2D_3D_callback(
    value: Optional[Tuple[int, int, int]]
) -> Optional[Tuple[int, ...]]:
    """
    Callback for options that require 2D or 3D inputs.

    In the case of 2D, the 3rd element should be set to -1.

    Parameters
    ----------
    value : (int, int, int)
        Tile size value.

    Returns
    -------
    (int, int, int) | (int, int)
        If the last element in `value` is -1 the tuple is reduced to the first two
        values.
    """
    if value is None:
        return value
    if value[2] == -1:
        return value[:2]
    return value



"""Random patching utilities."""

from typing import Generator, List, Optional, Tuple, Union

import numpy as np
import zarr

from .validate_patch_dimension import validate_patch_dimensions


# TOOD split in testable functions
def extract_patches_random(
    arr: np.ndarray,
    patch_size: Union[List[int], Tuple[int, ...]],
    target: Optional[np.ndarray] = None,
    seed: Optional[int] = None,
) -> Generator[Tuple[np.ndarray, Optional[np.ndarray]], None, None]:
    """
    Generate patches from an array in a random manner.

    The method calculates how many patches the image can be divided into and then
    extracts an equal number of random patches.

    It returns a generator that yields the following:

    - patch: np.ndarray, dimension C(Z)YX.
    - target_patch: np.ndarray, dimension C(Z)YX, if the target is present, None
        otherwise.

    Parameters
    ----------
    arr : np.ndarray
        Input image array.
    patch_size : Tuple[int]
        Patch sizes in each dimension.
    target : Optional[np.ndarray], optional
        Target array, by default None.
    seed : Optional[int], optional
        Random seed, by default None.

    Yields
    ------
    Generator[np.ndarray, None, None]
        Generator of patches.
    """
    rng = np.random.default_rng(seed=seed)

    is_3d_patch = len(patch_size) == 3

    # patches sanity check
    validate_patch_dimensions(arr, patch_size, is_3d_patch)

    # Update patch size to encompass S and C dimensions
    patch_size = [1, arr.shape[1], *patch_size]

    # iterate over the number of samples (S or T)
    for sample_idx in range(arr.shape[0]):
        # get sample array
        sample: np.ndarray = arr[sample_idx, ...]

        # same for target
        if target is not None:
            target_sample: np.ndarray = target[sample_idx, ...]

        # calculate the number of patches
        n_patches = np.ceil(np.prod(sample.shape) / np.prod(patch_size)).astype(int)

        # iterate over the number of patches
        for _ in range(n_patches):
            # get crop coordinates
            crop_coords = [
                rng.integers(0, sample.shape[i] - patch_size[1:][i], endpoint=True)
                for i in range(len(patch_size[1:]))
            ]

            # extract patch
            patch = (
                sample[
                    (
                        ...,  # type: ignore
                        *[  # type: ignore
                            slice(c, c + patch_size[1:][i])
                            for i, c in enumerate(crop_coords)
                        ],
                    )
                ]
                .copy()
                .astype(np.float32)
            )

            # same for target
            if target is not None:
                target_patch = (
                    target_sample[
                        (
                            ...,  # type: ignore
                            *[  # type: ignore
                                slice(c, c + patch_size[1:][i])
                                for i, c in enumerate(crop_coords)
                            ],
                        )
                    ]
                    .copy()
                    .astype(np.float32)
                )
                # return patch and target patch
                yield patch, target_patch
            else:
                # return patch
                yield patch, None


def extract_patches_random_from_chunks(
    arr: zarr.Array,
    patch_size: Union[List[int], Tuple[int, ...]],
    chunk_size: Union[List[int], Tuple[int, ...]],
    chunk_limit: Optional[int] = None,
    seed: Optional[int] = None,
) -> Generator[np.ndarray, None, None]:
    """
    Generate patches from an array in a random manner.

    The method calculates how many patches the image can be divided into and then
    extracts an equal number of random patches.

    Parameters
    ----------
    arr : np.ndarray
        Input image array.
    patch_size : Union[List[int], Tuple[int, ...]]
        Patch sizes in each dimension.
    chunk_size : Union[List[int], Tuple[int, ...]]
        Chunk sizes to load from the.
    chunk_limit : Optional[int], optional
        Number of chunks to load, by default None.
    seed : Optional[int], optional
        Random seed, by default None.

    Yields
    ------
    Generator[np.ndarray, None, None]
        Generator of patches.
    """
    is_3d_patch = len(patch_size) == 3

    # Patches sanity check
    validate_patch_dimensions(arr, patch_size, is_3d_patch)

    rng = np.random.default_rng(seed=seed)
    num_chunks = chunk_limit if chunk_limit else np.prod(arr._cdata_shape)

    # Iterate over num chunks in the array
    for _ in range(num_chunks):
        chunk_crop_coords = [
            rng.integers(0, max(0, arr.shape[i] - chunk_size[i]), endpoint=True)
            for i in range(len(chunk_size))
        ]
        chunk = arr[
            (
                ...,
                *[slice(c, c + chunk_size[i]) for i, c in enumerate(chunk_crop_coords)],
            )
        ].squeeze()

        # Add a singleton dimension if the chunk does not have a sample dimension
        if len(chunk.shape) == len(patch_size):
            chunk = np.expand_dims(chunk, axis=0)

        # Iterate over num samples (S)
        for sample_idx in range(chunk.shape[0]):
            spatial_chunk = chunk[sample_idx]
            assert len(spatial_chunk.shape) == len(
                patch_size
            ), "Requested chunk shape is not equal to patch size"

            n_patches = np.ceil(
                np.prod(spatial_chunk.shape) / np.prod(patch_size)
            ).astype(int)

            # Iterate over the number of patches
            for _ in range(n_patches):
                patch_crop_coords = [
                    rng.integers(
                        0, spatial_chunk.shape[i] - patch_size[i], endpoint=True
                    )
                    for i in range(len(patch_size))
                ]
                patch = (
                    spatial_chunk[
                        (
                            ...,
                            *[
                                slice(c, c + patch_size[i])
                                for i, c in enumerate(patch_crop_coords)
                            ],
                        )
                    ]
                    .copy()
                    .astype(np.float32)
                )
                yield patch



from torch import optim

from careamics.utils.torch_utils import get_optimizers, get_schedulers


def test_get_schedulers_exist():
    """Test that the function `get_schedulers` return
    existing torch schedulers.
    """
    for scheduler in get_schedulers():
        assert hasattr(optim.lr_scheduler, scheduler)


def test_get_optimizers_exist():
    """Test that the function `get_optimizers` return
    existing torch optimizers.
    """
    for optimizer in get_optimizers():
        assert hasattr(optim, optimizer)



"""Custom architecture Pydantic model."""

from __future__ import annotations

import inspect
from pprint import pformat
from typing import Any, Literal

from pydantic import ConfigDict, field_validator, model_validator
from torch.nn import Module
from typing_extensions import Self

from .architecture_model import ArchitectureModel
from .register_model import get_custom_model


class CustomModel(ArchitectureModel):
    """Custom model configuration.

    This Pydantic model allows storing parameters for a custom model. In order for the
    model to be valid, the specific model needs to be registered using the
    `register_model` decorator, and its name correctly passed to this model
    configuration (see Examples).

    Attributes
    ----------
    architecture : Literal["custom"]
        Discriminator for the custom model, must be set to "custom".
    name : str
        Name of the custom model.
    parameters : CustomParametersModel
        All parameters, required for the initialization of the torch module have to be
        passed here.

    Raises
    ------
    ValueError
        If the custom model `name` is unknown.
    ValueError
        If the custom model is not a torch Module subclass.
    ValueError
        If the custom model parameters are not valid.

    Examples
    --------
    >>> from torch import nn, ones
    >>> from careamics.config import CustomModel, register_model
    >>> # Register a custom model
    >>> @register_model(name="my_linear")
    ... class LinearModel(nn.Module):
    ...    def __init__(self, in_features, out_features, *args, **kwargs):
    ...        super().__init__()
    ...        self.in_features = in_features
    ...        self.out_features = out_features
    ...        self.weight = nn.Parameter(ones(in_features, out_features))
    ...        self.bias = nn.Parameter(ones(out_features))
    ...    def forward(self, input):
    ...        return (input @ self.weight) + self.bias
    ...
    >>> # Create a configuration
    >>> config_dict = {
    ...     "architecture": "custom",
    ...     "name": "my_linear",
    ...     "in_features": 10,
    ...     "out_features": 5,
    ... }
    >>> config = CustomModel(**config_dict)
    """

    # pydantic model config
    model_config = ConfigDict(
        extra="allow",
    )

    # discriminator used for choosing the pydantic model in Model
    architecture: Literal["custom"]
    """Name of the architecture."""

    name: str
    """Name of the custom model."""

    @field_validator("name")
    @classmethod
    def custom_model_is_known(cls, value: str) -> str:
        """Check whether the custom model is known.

        Parameters
        ----------
        value : str
            Name of the custom model as registered using the `@register_model`
            decorator.

        Returns
        -------
        str
            The custom model name.
        """
        # delegate error to get_custom_model
        model = get_custom_model(value)

        # check if it is a torch Module subclass
        if not issubclass(model, Module):
            raise ValueError(
                f'Retrieved class {model} with name "{value}" is not a '
                f"torch.nn.Module subclass."
            )

        return value

    @model_validator(mode="after")
    def check_parameters(self: Self) -> Self:
        """Validate model by instantiating the model with the parameters.

        Returns
        -------
        Self
            The validated model.
        """
        # instantiate model
        try:
            get_custom_model(self.name)(**self.model_dump())
        except Exception as e:
            raise ValueError(
                f"while passing parameters to the model {e}. Verify that all "
                f"mandatory parameters are provided, and that either the {e} accepts "
                f"*args and **kwargs in its __init__() method, or that no additional"
                f"parameter is provided. Trace: "
                f"filename: {inspect.trace()[-1].filename}, function: "
                f"{inspect.trace()[-1].function}, line: {inspect.trace()[-1].lineno}"
            ) from None

        return self

    def __str__(self) -> str:
        """Pretty string representing the configuration.

        Returns
        -------
        str
            Pretty string.
        """
        return pformat(self.model_dump())

    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        """Dump the model configuration.

        Parameters
        ----------
        **kwargs : Any
            Additional keyword arguments from Pydantic BaseModel model_dump method.

        Returns
        -------
        dict[str, Any]
            Model configuration.
        """
        model_dict = super().model_dump()

        # remove the name key
        model_dict.pop("name")

        return model_dict



"""
Lightning Module for LadderVAE.
"""

from typing import Any, Dict

import ml_collections
import numpy as np
import pytorch_lightning as L
import torch
import torchvision.transforms.functional as F

from careamics.models.lvae.likelihoods import LikelihoodModule
from careamics.models.lvae.lvae import LadderVAE
from careamics.models.lvae.utils import (
    LossType,
    compute_batch_mean,
    free_bits_kl,
    torch_nanmean,
)

from .metrics import RangeInvariantPsnr, RunningPSNR
from .train_utils import MetricMonitor


class LadderVAELight(L.LightningModule):

    def __init__(
        self,
        config: ml_collections.ConfigDict,
        data_mean: Dict[str, torch.Tensor],
        data_std: Dict[str, torch.Tensor],
        target_ch: int,
    ):
        """
        Here we will do the following:
            - initialize the model (from LadderVAE class)
            - initialize the parameters related to the training and loss.

        NOTE:
        Some of the model attributes are defined in the model object itself, while some others will be defined here.
        Note that all the attributes related to the training and loss that were already defined in the model object
        are redefined here as Lightning module attributes (e.g., self.some_attr = model.some_attr).
        The attributes related to the model itself are treated as model attributes (e.g., self.model.some_attr).

        NOTE: HC stands for Hard Coded attribute.
        """
        super().__init__()

        self.data_mean = data_mean
        self.data_std = data_std
        self.target_ch = target_ch

        # Initialize LVAE model
        self.model = LadderVAE(
            data_mean=data_mean, data_std=data_std, config=config, target_ch=target_ch
        )

        ##### Define attributes from config #####
        self.workdir = config.workdir
        self._input_is_sum = False
        self.kl_loss_formulation = config.loss.kl_loss_formulation
        assert self.kl_loss_formulation in [
            None,
            "",
            "usplit",
            "denoisplit",
            "denoisplit_usplit",
        ], f"""
            Invalid kl_loss_formulation. {self.kl_loss_formulation}"""

        ##### Define loss attributes #####
        # Parameters already defined in the model object
        self.loss_type = self.model.loss_type
        self._denoisplit_w = self._usplit_w = None
        if self.loss_type == LossType.DenoiSplitMuSplit:
            self._usplit_w = 0
            self._denoisplit_w = 1 - self._usplit_w
            assert self._denoisplit_w + self._usplit_w == 1
        self._restricted_kl = self.model._restricted_kl

        # General loss parameters
        self.channel_1_w = 1
        self.channel_2_w = 1

        # About Reconsruction Loss
        self.reconstruction_mode = False
        self.skip_nboundary_pixels_from_loss = None
        self.reconstruction_weight = 1.0
        self._exclusion_loss_weight = 0
        self.ch1_recons_w = 1
        self.ch2_recons_w = 1
        self.enable_mixed_rec = False
        self.mixed_rec_w_step = 0

        # About KL Loss
        self.kl_weight = 1.0  # HC
        self.usplit_kl_weight = None  # HC
        self.free_bits = 1.0  # HC
        self.kl_annealing = False  # HC
        self.kl_annealtime = self.kl_start = None
        if self.kl_annealing:
            self.kl_annealtime = 10  # HC
            self.kl_start = -1  # HC

        ##### Define training attributes #####
        self.lr = config.training.lr
        self.lr_scheduler_patience = config.training.lr_scheduler_patience
        self.lr_scheduler_monitor = config.model.get("monitor", "val_loss")
        self.lr_scheduler_mode = MetricMonitor(self.lr_scheduler_monitor).mode()

        # Initialize object for keeping track of PSNR for each output channel
        self.channels_psnr = [RunningPSNR() for _ in range(self.model.target_ch)]

    def forward(self, x: Any) -> Any:
        return self.model(x)

    def training_step(
        self, batch: torch.Tensor, batch_idx: int, enable_logging: bool = True
    ) -> Dict[str, torch.Tensor]:

        if self.current_epoch == 0 and batch_idx == 0:
            self.log("val_psnr", 1.0, on_epoch=True)

        # Pre-processing of inputs
        x, target = batch[:2]
        self.set_params_to_same_device_as(x)
        x_normalized = self.normalize_input(x)
        if self.reconstruction_mode:  # just for experimental purpose
            target_normalized = x_normalized[:, :1].repeat(1, 2, 1, 1)
            target = None
            mask = None
        else:
            target_normalized = self.normalize_target(target)
            mask = ~((target == 0).reshape(len(target), -1).all(dim=1))

        # Forward pass
        out, td_data = self.forward(x_normalized)

        if (
            self.model.encoder_no_padding_mode
            and out.shape[-2:] != target_normalized.shape[-2:]
        ):
            target_normalized = F.center_crop(target_normalized, out.shape[-2:])

        # Loss Computations
        # mask = torch.isnan(target.reshape(len(x), -1)).all(dim=1)
        recons_loss_dict, imgs = self.get_reconstruction_loss(
            reconstruction=out,
            target=target_normalized,
            input=x_normalized,
            splitting_mask=mask,
            return_predicted_img=True,
        )

        # This `if` is not used by default config
        if self.skip_nboundary_pixels_from_loss:
            pad = self.skip_nboundary_pixels_from_loss
            target_normalized = target_normalized[:, :, pad:-pad, pad:-pad]

        recons_loss = recons_loss_dict["loss"] * self.reconstruction_weight

        if torch.isnan(recons_loss).any():
            recons_loss = 0.0

        if self.model.non_stochastic_version:
            kl_loss = torch.Tensor([0.0]).cuda()
            net_loss = recons_loss
        else:
            if self.loss_type == LossType.DenoiSplitMuSplit:
                msg = f"For the loss type {LossType.name(self.loss_type)}, kl_loss_formulation must be denoisplit_usplit"
                assert self.kl_loss_formulation == "denoisplit_usplit", msg
                assert self._denoisplit_w is not None and self._usplit_w is not None

                kl_key_denoisplit = "kl_restricted" if self._restricted_kl else "kl"
                # NOTE: 'kl' key stands for the 'kl_samplewise' key in the TopDownLayer class.
                # The different naming comes from `top_down_pass()` method in the LadderVAE class.
                denoisplit_kl = self.get_kl_divergence_loss(
                    topdown_layer_data_dict=td_data, kl_key=kl_key_denoisplit
                )
                usplit_kl = self.get_kl_divergence_loss_usplit(
                    topdown_layer_data_dict=td_data
                )
                kl_loss = (
                    self._denoisplit_w * denoisplit_kl + self._usplit_w * usplit_kl
                )
                kl_loss = self.kl_weight * kl_loss

                recons_loss = self.reconstruction_loss_musplit_denoisplit(
                    out, target_normalized
                )
                # recons_loss = self._denoisplit_w * recons_loss_nm + self._usplit_w * recons_loss_gm

            elif self.kl_loss_formulation == "usplit":
                kl_loss = self.get_kl_weight() * self.get_kl_divergence_loss_usplit(
                    td_data
                )
            elif self.kl_loss_formulation in ["", "denoisplit"]:
                kl_loss = self.get_kl_weight() * self.get_kl_divergence_loss(td_data)
            net_loss = recons_loss + kl_loss

        # Logging
        if enable_logging:
            for i, x in enumerate(td_data["debug_qvar_max"]):
                self.log(f"qvar_max:{i}", x.item(), on_epoch=True)

            self.log("reconstruction_loss", recons_loss_dict["loss"], on_epoch=True)
            self.log("kl_loss", kl_loss, on_epoch=True)
            self.log("training_loss", net_loss, on_epoch=True)
            self.log("lr", self.lr, on_epoch=True)
            if self.model._tethered_ch2_scalar is not None:
                self.log(
                    "tethered_ch2_scalar",
                    self.model._tethered_ch2_scalar,
                    on_epoch=True,
                )
                self.log(
                    "tethered_ch1_scalar",
                    self.model._tethered_ch1_scalar,
                    on_epoch=True,
                )

            # self.log('grad_norm_bottom_up', self.grad_norm_bottom_up, on_epoch=True)
            # self.log('grad_norm_top_down', self.grad_norm_top_down, on_epoch=True)

        output = {
            "loss": net_loss,
            "reconstruction_loss": (
                recons_loss.detach()
                if isinstance(recons_loss, torch.Tensor)
                else recons_loss
            ),
            "kl_loss": kl_loss.detach(),
        }
        # https://github.com/openai/vdvae/blob/main/train.py#L26
        if torch.isnan(net_loss).any():
            return None

        return output

    def validation_step(self, batch: torch.Tensor, batch_idx: int):
        # Pre-processing of inputs
        x, target = batch[:2]
        self.set_params_to_same_device_as(x)
        x_normalized = self.normalize_input(x)
        if self.reconstruction_mode:  # only for experimental purpose
            target_normalized = x_normalized[:, :1].repeat(1, 2, 1, 1)
            target = None
            mask = None
        else:
            target_normalized = self.normalize_target(target)
            mask = ~((target == 0).reshape(len(target), -1).all(dim=1))

        # Forward pass
        out, _ = self.forward(x_normalized)

        if self.model.predict_logvar is not None:
            out_mean, _ = out.chunk(2, dim=1)
        else:
            out_mean = out

        if (
            self.model.encoder_no_padding_mode
            and out.shape[-2:] != target_normalized.shape[-2:]
        ):
            target_normalized = F.center_crop(target_normalized, out.shape[-2:])

        if self.loss_type == LossType.DenoiSplitMuSplit:
            recons_loss = self.reconstruction_loss_musplit_denoisplit(
                out, target_normalized
            )
            recons_loss_dict = {"loss": recons_loss}
            recons_img = out_mean
        else:
            # Metrics computation
            recons_loss_dict, recons_img = self.get_reconstruction_loss(
                reconstruction=out_mean,
                target=target_normalized,
                input=x_normalized,
                splitting_mask=mask,
                return_predicted_img=True,
            )

        # This `if` is not used by default config
        if self.skip_nboundary_pixels_from_loss:
            pad = self.skip_nboundary_pixels_from_loss
            target_normalized = target_normalized[:, :, pad:-pad, pad:-pad]

        channels_rinvpsnr = []
        for i in range(target_normalized.shape[1]):
            self.channels_psnr[i].update(recons_img[:, i], target_normalized[:, i])
            psnr = RangeInvariantPsnr(
                target_normalized[:, i].clone(), recons_img[:, i].clone()
            )
            channels_rinvpsnr.append(psnr)
            psnr = torch_nanmean(psnr).item()
            self.log(f"val_psnr_l{i+1}", psnr, on_epoch=True)

        recons_loss = recons_loss_dict["loss"]
        if torch.isnan(recons_loss).any():
            return

        self.log("val_loss", recons_loss, on_epoch=True)
        # self.log('val_psnr', (val_psnr_l1 + val_psnr_l2) / 2, on_epoch=True)

        # if batch_idx == 0 and self.power_of_2(self.current_epoch):
        #     all_samples = []
        #     for i in range(20):
        #         sample, _ = self(x_normalized[0:1, ...])
        #         sample = self.likelihood.get_mean_lv(sample)[0]
        #         all_samples.append(sample[None])

        #     all_samples = torch.cat(all_samples, dim=0)
        #     all_samples = all_samples * self.data_std + self.data_mean
        #     all_samples = all_samples.cpu()
        #     img_mmse = torch.mean(all_samples, dim=0)[0]
        #     self.log_images_for_tensorboard(all_samples[:, 0, 0, ...], target[0, 0, ...], img_mmse[0], 'label1')
        #     self.log_images_for_tensorboard(all_samples[:, 0, 1, ...], target[0, 1, ...], img_mmse[1], 'label2')

        # return net_loss

    def on_validation_epoch_end(self):
        psnr_arr = []
        for i in range(len(self.channels_psnr)):
            psnr = self.channels_psnr[i].get()
            if psnr is None:
                psnr_arr = None
                break
            psnr_arr.append(psnr.cpu().numpy())
            self.channels_psnr[i].reset()

        if psnr_arr is not None:
            psnr = np.mean(psnr_arr)
            self.log("val_psnr", psnr, on_epoch=True)
        else:
            self.log("val_psnr", 0.0, on_epoch=True)

        if self.mixed_rec_w_step:
            self.mixed_rec_w = max(self.mixed_rec_w - self.mixed_rec_w_step, 0.0)
            self.log("mixed_rec_w", self.mixed_rec_w, on_epoch=True)

    def predict_step(self, batch: torch.Tensor, batch_idx: Any) -> Any:
        raise NotImplementedError("predict_step is not implemented")

    def configure_optimizers(self):
        optimizer = torch.optim.Adamax(self.parameters(), lr=self.lr, weight_decay=0)
        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            optimizer,
            self.lr_scheduler_mode,
            patience=self.lr_scheduler_patience,
            factor=0.5,
            min_lr=1e-12,
            verbose=True,
        )

        return {
            "optimizer": optimizer,
            "lr_scheduler": scheduler,
            "monitor": self.lr_scheduler_monitor,
        }

    ##### REQUIRED Methods for Loss Computation #####
    def get_reconstruction_loss(
        self,
        reconstruction: torch.Tensor,
        target: torch.Tensor,
        input: torch.Tensor,
        splitting_mask: torch.Tensor = None,
        return_predicted_img: bool = False,
        likelihood_obj: LikelihoodModule = None,
    ) -> Dict[str, torch.Tensor]:
        """
        Parameters
        ----------
        reconstruction: torch.Tensor,
        target: torch.Tensor
        input: torch.Tensor
        splitting_mask: torch.Tensor = None
            A boolean tensor that indicates which items to keep for reconstruction loss computation.
            If `None`, all the elements of the items are considered (i.e., the mask is all `True`).
        return_predicted_img: bool = False
        likelihood_obj: LikelihoodModule = None
        """
        output = self._get_reconstruction_loss_vector(
            reconstruction=reconstruction,
            target=target,
            input=input,
            return_predicted_img=return_predicted_img,
            likelihood_obj=likelihood_obj,
        )
        loss_dict = output[0] if return_predicted_img else output

        if splitting_mask is None:
            splitting_mask = torch.ones_like(loss_dict["loss"]).bool()

        # print(len(target) - (torch.isnan(loss_dict['loss'])).sum())

        loss_dict["loss"] = loss_dict["loss"][splitting_mask].sum() / len(
            reconstruction
        )
        for i in range(1, 1 + target.shape[1]):
            key = f"ch{i}_loss"
            loss_dict[key] = loss_dict[key][splitting_mask].sum() / len(reconstruction)

        if "mixed_loss" in loss_dict:
            loss_dict["mixed_loss"] = torch.mean(loss_dict["mixed_loss"])
        if return_predicted_img:
            assert len(output) == 2
            return loss_dict, output[1]
        else:
            return loss_dict

    def _get_reconstruction_loss_vector(
        self,
        reconstruction: torch.Tensor,
        target: torch.Tensor,
        input: torch.Tensor,
        return_predicted_img: bool = False,
        likelihood_obj: LikelihoodModule = None,
    ):
        """
        Parameters
        ----------
        return_predicted_img: bool
            If set to `True`, the besides the loss, the reconstructed image is also returned.
            Default is `False`.
        """
        output = {
            "loss": None,
            "mixed_loss": None,
        }

        for i in range(1, 1 + target.shape[1]):
            output[f"ch{i}_loss"] = None

        if likelihood_obj is None:
            likelihood_obj = self.model.likelihood

        # Log likelihood
        ll, like_dict = likelihood_obj(reconstruction, target)
        ll = self._get_weighted_likelihood(ll)
        if (
            self.skip_nboundary_pixels_from_loss is not None
            and self.skip_nboundary_pixels_from_loss > 0
        ):
            pad = self.skip_nboundary_pixels_from_loss
            ll = ll[:, :, pad:-pad, pad:-pad]
            like_dict["params"]["mean"] = like_dict["params"]["mean"][
                :, :, pad:-pad, pad:-pad
            ]

        # assert ll.shape[1] == 2, f"Change the code below to handle >2 channels first. ll.shape {ll.shape}"
        output = {"loss": compute_batch_mean(-1 * ll)}
        if ll.shape[1] > 1:
            for i in range(1, 1 + target.shape[1]):
                output[f"ch{i}_loss"] = compute_batch_mean(-ll[:, i - 1])
        else:
            assert ll.shape[1] == 1
            output["ch1_loss"] = output["loss"]
            output["ch2_loss"] = output["loss"]

        if (
            self.channel_1_w is not None
            and self.channel_2_w is not None
            and (self.channel_1_w != 1 or self.channel_2_w != 1)
        ):
            assert ll.shape[1] == 2, "Only 2 channels are supported for now."
            output["loss"] = (
                self.channel_1_w * output["ch1_loss"]
                + self.channel_2_w * output["ch2_loss"]
            ) / (self.channel_1_w + self.channel_2_w)

        # This `if` is not used by default config
        if self.enable_mixed_rec:
            mixed_pred, mixed_logvar = self.get_mixed_prediction(
                like_dict["params"]["mean"],
                like_dict["params"]["logvar"],
                self.data_mean,
                self.data_std,
            )
            if (
                self.model._multiscale_count is not None
                and self.model._multiscale_count > 1
            ):
                assert input.shape[1] == self.model._multiscale_count
                input = input[:, :1]

            assert (
                input.shape == mixed_pred.shape
            ), "No fucking room for vectorization induced bugs."
            mixed_recons_ll = self.model.likelihood.log_likelihood(
                input, {"mean": mixed_pred, "logvar": mixed_logvar}
            )
            output["mixed_loss"] = compute_batch_mean(-1 * mixed_recons_ll)

        # This `if` is not used by default config
        if self._exclusion_loss_weight:
            raise NotImplementedError(
                "Exclusion loss is not well defined here, so it should not be used."
            )
            imgs = like_dict["params"]["mean"]
            exclusion_loss = compute_exclusion_loss(imgs[:, :1], imgs[:, 1:])
            output["exclusion_loss"] = exclusion_loss

        if return_predicted_img:
            return output, like_dict["params"]["mean"]

        return output

    def reconstruction_loss_musplit_denoisplit(self, out, target_normalized):
        if self.model.predict_logvar is not None:
            out_mean, _ = out.chunk(2, dim=1)
        else:
            out_mean = out

        recons_loss_nm = (
            -1 * self.model.likelihood_NM(out_mean, target_normalized)[0].mean()
        )
        recons_loss_gm = -1 * self.model.likelihood_gm(out, target_normalized)[0].mean()
        recons_loss = (
            self._denoisplit_w * recons_loss_nm + self._usplit_w * recons_loss_gm
        )
        return recons_loss

    def _get_weighted_likelihood(self, ll):
        """
        Each of the channels gets multiplied with a different weight.
        """
        if self.ch1_recons_w == 1 and self.ch2_recons_w == 1:
            return ll

        assert ll.shape[1] == 2, "This function is only for 2 channel images"

        mask1 = torch.zeros((len(ll), ll.shape[1], 1, 1), device=ll.device)
        mask1[:, 0] = 1
        mask2 = torch.zeros((len(ll), ll.shape[1], 1, 1), device=ll.device)
        mask2[:, 1] = 1

        return ll * mask1 * self.ch1_recons_w + ll * mask2 * self.ch2_recons_w

    def get_kl_weight(self):
        """
        KL loss can be weighted depending whether any annealing procedure is used.
        This function computes the weight of the KL loss in case of annealing.
        """
        if self.kl_annealing == True:
            # calculate relative weight
            kl_weight = (self.current_epoch - self.kl_start) * (
                1.0 / self.kl_annealtime
            )
            # clamp to [0,1]
            kl_weight = min(max(0.0, kl_weight), 1.0)

            # if the final weight is given, then apply that weight on top of it
            if self.kl_weight is not None:
                kl_weight = kl_weight * self.kl_weight
        elif self.kl_weight is not None:
            return self.kl_weight
        else:
            kl_weight = 1.0
        return kl_weight

    def get_kl_divergence_loss_usplit(
        self, topdown_layer_data_dict: Dict[str, torch.Tensor]
    ) -> torch.Tensor:
        """ """
        kl = torch.cat(
            [kl_layer.unsqueeze(1) for kl_layer in topdown_layer_data_dict["kl"]], dim=1
        )
        # NOTE: kl.shape = (16,4) 16 is batch size. 4 is number of layers.
        # Values are sum() and so are of the order 30000
        # Example values: 30626.6758, 31028.8145, 29509.8809, 29945.4922, 28919.1875, 29075.2988

        nlayers = kl.shape[1]
        for i in range(nlayers):
            # topdown_layer_data_dict['z'][2].shape[-3:] = 128 * 32 * 32
            norm_factor = np.prod(topdown_layer_data_dict["z"][i].shape[-3:])
            # if self._restricted_kl:
            #     pow = np.power(2,min(i + 1, self._multiscale_count-1))
            #     norm_factor /= pow * pow

            kl[:, i] = kl[:, i] / norm_factor

        kl_loss = free_bits_kl(kl, 0.0).mean()
        return kl_loss

    def get_kl_divergence_loss(self, topdown_layer_data_dict, kl_key="kl"):
        """
        kl[i] for each i has length batch_size
        resulting kl shape: (batch_size, layers)
        """
        kl = torch.cat(
            [kl_layer.unsqueeze(1) for kl_layer in topdown_layer_data_dict[kl_key]],
            dim=1,
        )

        # As compared to uSplit kl divergence,
        # more by a factor of 4 just because we do sum and not mean.
        kl_loss = free_bits_kl(kl, self.free_bits).sum()
        # NOTE: at each hierarchy, it is more by a factor of 128/i**2).
        # 128/(2*2) = 32 (bottommost layer)
        # 128/(4*4) = 8
        # 128/(8*8) = 2
        # 128/(16*16) = 0.5 (topmost layer)

        # Normalize the KL-loss w.r.t. the  latent space
        kl_loss = kl_loss / np.prod(self.model.img_shape)
        return kl_loss

    ##### UTILS Methods #####
    def normalize_input(self, x):
        if self.model.normalized_input:
            return x
        return (x - self.data_mean["input"].mean()) / self.data_std["input"].mean()

    def normalize_target(self, target, batch=None):
        return (target - self.data_mean["target"]) / self.data_std["target"]

    def unnormalize_target(self, target_normalized):
        return target_normalized * self.data_std["target"] + self.data_mean["target"]

    ##### ADDITIONAL Methods #####
    # def log_images_for_tensorboard(self, pred, target, img_mmse, label):
    #     clamped_pred = torch.clamp((pred - pred.min()) / (pred.max() - pred.min()), 0, 1)
    #     clamped_mmse = torch.clamp((img_mmse - img_mmse.min()) / (img_mmse.max() - img_mmse.min()), 0, 1)
    #     if target is not None:
    #         clamped_input = torch.clamp((target - target.min()) / (target.max() - target.min()), 0, 1)
    #         img = wandb.Image(clamped_input[None].cpu().numpy())
    #         self.logger.experiment.log({f'target_for{label}': img})
    #         # self.trainer.logger.experiment.add_image(f'target_for{label}', clamped_input[None], self.current_epoch)
    #     for i in range(3):
    #         # self.trainer.logger.experiment.add_image(f'{label}/sample_{i}', clamped_pred[i:i + 1], self.current_epoch)
    #         img = wandb.Image(clamped_pred[i:i + 1].cpu().numpy())
    #         self.logger.experiment.log({f'{label}/sample_{i}': img})

    #     img = wandb.Image(clamped_mmse[None].cpu().numpy())
    #     self.trainer.logger.experiment.log({f'{label}/mmse (100 samples)': img})

    @property
    def global_step(self) -> int:
        """Global step."""
        return self._global_step

    def increment_global_step(self):
        """Increments global step by 1."""
        self._global_step += 1

    def set_params_to_same_device_as(self, correct_device_tensor: torch.Tensor):

        self.model.likelihood.set_params_to_same_device_as(correct_device_tensor)
        if isinstance(self.data_mean, torch.Tensor):
            if self.data_mean.device != correct_device_tensor.device:
                self.data_mean = self.data_mean.to(correct_device_tensor.device)
                self.data_std = self.data_std.to(correct_device_tensor.device)
        elif isinstance(self.data_mean, dict):
            for k, v in self.data_mean.items():
                if v.device != correct_device_tensor.device:
                    self.data_mean[k] = v.to(correct_device_tensor.device)
                    self.data_std[k] = self.data_std[k].to(correct_device_tensor.device)

    def get_mixed_prediction(
        self, prediction, prediction_logvar, data_mean, data_std, channel_weights=None
    ):
        pred_unorm = prediction * data_std["target"] + data_mean["target"]
        if channel_weights is None:
            channel_weights = 1

        if self._input_is_sum:
            mixed_prediction = torch.sum(
                pred_unorm * channel_weights, dim=1, keepdim=True
            )
        else:
            mixed_prediction = torch.mean(
                pred_unorm * channel_weights, dim=1, keepdim=True
            )

        mixed_prediction = (mixed_prediction - data_mean["input"].mean()) / data_std[
            "input"
        ].mean()

        if prediction_logvar is not None:
            if data_std["target"].shape == data_std["input"].shape and torch.all(
                data_std["target"] == data_std["input"]
            ):
                assert channel_weights == 1
                logvar = prediction_logvar
            else:
                var = torch.exp(prediction_logvar)
                var = var * (data_std["target"] / data_std["input"]) ** 2
                if channel_weights != 1:
                    var = var * torch.square(channel_weights)

                # sum of variance.
                mixed_var = 0
                for i in range(var.shape[1]):
                    mixed_var += var[:, i : i + 1]

                logvar = torch.log(mixed_var)
        else:
            logvar = None
        return mixed_prediction, logvar



"""Bioimage.io utils."""

from pathlib import Path
from typing import Union


def get_unzip_path(zip_path: Union[Path, str]) -> Path:
    """Generate unzipped folder path from the bioimage.io model path.

    Parameters
    ----------
    zip_path : Path
        Path to the bioimage.io model.

    Returns
    -------
    Path
        Path to the unzipped folder.
    """
    zip_path = Path(zip_path)

    return zip_path.parent / (str(zip_path.name) + ".unzip")


def create_env_text(pytorch_version: str, torchvision_version: str) -> str:
    """Create environment yaml content for the bioimage model.

    This installs an environment with the specified pytorch version and the latest
    changes to careamics.

    Parameters
    ----------
    pytorch_version : str
        Pytorch version.
    torchvision_version : str
        Torchvision version.

    Returns
    -------
    str
        Environment text.
    """
    env = (
        f"name: careamics\n"
        f"dependencies:\n"
        f"  - python=3.10\n"
        f"  - pytorch={pytorch_version}\n"
        f"  - torchvision={torchvision_version}\n"
        f"  - pip\n"
        f"  - pip:\n"
        f"    - git+https://github.com/CAREamics/careamics.git\n"
    )

    return env



"""
Context submodule.

A convenience function to change the working directory in order to save data.
"""

import os
from contextlib import contextmanager
from pathlib import Path
from typing import Iterator, Union


def get_careamics_home() -> Path:
    """Return the CAREamics home directory.

    CAREamics home directory is a hidden folder in home.

    Returns
    -------
    Path
        CAREamics home directory path.
    """
    home = Path.home() / ".careamics"

    if not home.exists():
        home.mkdir(parents=True, exist_ok=True)

    return home


@contextmanager
def cwd(path: Union[str, Path]) -> Iterator[None]:
    """
    Change the current working directory to the given path.

    This method can be used to generate files in a specific directory, once out of the
    context, the working directory is set back to the original one.

    Parameters
    ----------
    path : Union[str,Path]
        New working directory path.

    Returns
    -------
    Iterator[None]
        None values.

    Examples
    --------
    The context is whcnaged within the block and then restored to the original one.

    >>> with cwd(my_path):
    ...     pass # do something
    """
    path = Path(path)

    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)

    old_pwd = Path(".").absolute()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_pwd)



"""Test-time augmentations."""

from torch import Tensor, flip, mean, rot90, stack


class ImageRestorationTTA:
    """
    Test-time augmentation for image restoration tasks.

    The augmentation is performed using all 90 deg rotations and their flipped version,
    as well as the original image flipped.

    Tensors should be of shape SC(Z)YX.

    This transformation is used in the LightningModule in order to perform test-time
    augmentation.
    """

    def forward(self, input_tensor: Tensor) -> list[Tensor]:
        """
        Apply test-time augmentation to the input tensor.

        Parameters
        ----------
        input_tensor : Tensor
            Input tensor, shape SC(Z)YX.

        Returns
        -------
        list of torch.Tensor
            List of augmented tensors.
        """
        # axes: only applies to YX axes
        axes = (-2, -1)

        augmented = [
            # original
            input_tensor,
            # rotations
            rot90(input_tensor, 1, dims=axes),
            rot90(input_tensor, 2, dims=axes),
            rot90(input_tensor, 3, dims=axes),
            # original flipped
            flip(input_tensor, dims=(axes[0],)),
            flip(input_tensor, dims=(axes[1],)),
        ]

        # rotated once, flipped
        augmented.extend(
            [
                flip(augmented[1], dims=(axes[0],)),
                flip(augmented[1], dims=(axes[1],)),
            ]
        )

        return augmented

    def backward(self, x: list[Tensor]) -> Tensor:
        """Undo the test-time augmentation.

        Parameters
        ----------
        x : Any
            List of augmented tensors of shape SC(Z)YX.

        Returns
        -------
        Any
            Original tensor.
        """
        axes = (-2, -1)

        reverse = [
            # original
            x[0],
            # rotated
            rot90(x[1], -1, dims=axes),
            rot90(x[2], -2, dims=axes),
            rot90(x[3], -3, dims=axes),
            # original flipped
            flip(x[4], dims=(axes[0],)),
            flip(x[5], dims=(axes[1],)),
            # rotated once, flipped
            rot90(flip(x[6], dims=(axes[0],)), -1, dims=axes),
            rot90(flip(x[7], dims=(axes[1],)), -1, dims=axes),
        ]

        return mean(stack(reverse), dim=0)



"""Noise models config."""

from pathlib import Path
from typing import Literal, Optional, Union

import numpy as np
import torch
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PlainSerializer,
    PlainValidator,
)
from typing_extensions import Annotated

from careamics.utils.serializers import _array_to_json, _to_numpy

# TODO: this is a temporary solution to serialize and deserialize array fields
# in pydantic models. Specifically, the aim is to enable saving and loading configs
# with such arrays to/from JSON files during, resp., training and evaluation.
Array = Annotated[
    Union[np.ndarray, torch.Tensor],
    PlainSerializer(_array_to_json, return_type=str),
    PlainValidator(_to_numpy),
]
"""Annotated array type, used to serialize arrays or tensors to JSON strings
and deserialize them back to arrays."""


# TODO: add histogram-based noise model


class GaussianMixtureNMConfig(BaseModel):
    """Gaussian mixture noise model."""

    model_config = ConfigDict(
        protected_namespaces=(),
        validate_assignment=True,
        arbitrary_types_allowed=True,
        extra="allow",
    )
    # model type
    model_type: Literal["GaussianMixtureNoiseModel"]

    path: Optional[Union[Path, str]] = None
    """Path to the directory where the trained noise model (*.npz) is saved in the
    `train` method."""

    # TODO remove and use as parameters to the NM functions?
    signal: Optional[Union[str, Path, np.ndarray]] = Field(default=None, exclude=True)
    """Path to the file containing signal or respective numpy array."""

    # TODO remove and use as parameters to the NM functions?
    observation: Optional[Union[str, Path, np.ndarray]] = Field(
        default=None, exclude=True
    )
    """Path to the file containing observation or respective numpy array."""

    weight: Optional[Array] = None
    """A [3*n_gaussian, n_coeff] sized array containing the values of the weights
    describing the GMM noise model, with each row corresponding to one
    parameter of each gaussian, namely [mean, standard deviation and weight].
    Specifically, rows are organized as follows:
    - first n_gaussian rows correspond to the means
    - next n_gaussian rows correspond to the weights
    - last n_gaussian rows correspond to the standard deviations
    If `weight=None`, the weight array is initialized using the `min_signal`
    and `max_signal` parameters."""

    n_gaussian: int = Field(default=1, ge=1)
    """Number of gaussians used for the GMM."""

    n_coeff: int = Field(default=2, ge=2)
    """Number of coefficients to describe the functional relationship between gaussian
    parameters and the signal. 2 implies a linear relationship, 3 implies a quadratic
    relationship and so on."""

    min_signal: float = Field(default=0.0, ge=0.0)
    """Minimum signal intensity expected in the image."""

    max_signal: float = Field(default=1.0, ge=0.0)
    """Maximum signal intensity expected in the image."""

    min_sigma: float = Field(default=200.0, ge=0.0)  # TODO took from nb in pn2v
    """Minimum value of `standard deviation` allowed in the GMM.
    All values of `standard deviation` below this are clamped to this value."""

    tol: float = Field(default=1e-10)
    """Tolerance used in the computation of the noise model likelihood."""

    # @model_validator(mode="after")
    # def validate_path_to_pretrained_vs_training_data(self: Self) -> Self:
    #     """Validate paths provided in the config.

    #     Returns
    #     -------
    #     Self
    #         Returns itself.
    #     """
    #     if self.path and (self.signal is not None or self.observation is not None):
    #         raise ValueError(
    #             "Either only 'path' to pre-trained noise model should be"
    #             "provided or only signal and observation in form of paths"
    #             "or numpy arrays."
    #         )
    #     if not self.path and (self.signal is None or self.observation is None):
    #         raise ValueError(
    #             "Either only 'path' to pre-trained noise model should be"
    #             "provided or only signal and observation in form of paths"
    #             "or numpy arrays."
    #         )
    #     return self
    # TODO revisit validation


# The noise model is given by a set of GMMs, one for each target
# e.g., 2 target channels, 2 noise models
class MultiChannelNMConfig(BaseModel):
    """Noise Model config aggregating noise models for single output channels."""

    # TODO: check that this model config is OK
    model_config = ConfigDict(
        validate_assignment=True, arbitrary_types_allowed=True, extra="allow"
    )
    noise_models: list[GaussianMixtureNMConfig]
    """List of noise models, one for each target channel."""



from careamics.config.transformations import XYFlipModel
from careamics.transforms import XYFlip


def test_comptatibility_with_transform():
    """Test that the model allows instantiating a transform."""
    model = XYFlipModel(flip_y=False, p=0.3)

    # instantiate transform
    XYFlip(**model.model_dump())



"""StructN2V axes supported by CAREamics."""

from careamics.utils import BaseEnum


class SupportedStructAxis(str, BaseEnum):
    """Supported structN2V mask axes.

    Attributes
    ----------
    HORIZONTAL : str
        Horizontal axis.
    VERTICAL : str
        Vertical axis.
    NONE : str
        No axis, the mask is not applied.
    """

    HORIZONTAL = "horizontal"
    VERTICAL = "vertical"
    NONE = "none"
